@@ -1,5 +1,6 @@
 # frozen_string_literal: true
 require 'capybara/session/matchers'
+require 'addressable/uri'
 
 module Capybara
 
@@ -20,7 +21,7 @@ module Capybara
   # +current_path, and so on. It also delegate a number of methods to a Capybara::Document, representing
   # the current HTML document. This allows interaction:
   #
-  #     session.fill_in('q', :with => 'Capybara')
+  #     session.fill_in('q', with: 'Capybara')
   #     session.click_button('Search')
   #     expect(session).to have_content('Capybara')
   #
@@ -40,6 +41,7 @@ module Capybara
       :has_no_table?, :has_table?, :unselect, :has_select?, :has_no_select?,
       :has_selector?, :has_no_selector?, :click_on, :has_no_checked_field?,
       :has_no_unchecked_field?, :query, :assert_selector, :assert_no_selector,
+      :assert_all_of_selectors, :assert_none_of_selectors,
       :refute_selector, :assert_text, :assert_no_text
     ]
     # @api private
@@ -49,8 +51,8 @@ module Capybara
     SESSION_METHODS = [
       :body, :html, :source, :current_url, :current_host, :current_path,
       :execute_script, :evaluate_script, :visit, :go_back, :go_forward,
-      :within, :within_fieldset, :within_table, :within_frame, :current_window,
-      :windows, :open_new_window, :switch_to_window, :within_window, :window_opened_by,
+      :within, :within_element, :within_fieldset, :within_table, :within_frame, :switch_to_frame,
+      :current_window, :windows, :open_new_window, :switch_to_window, :within_window, :window_opened_by,
       :save_page, :save_and_open_page, :save_screenshot,
       :save_and_open_screenshot, :reset_session!, :response_headers,
       :status_code, :current_scope,
@@ -66,6 +68,7 @@ module Capybara
     attr_accessor :synchronized
 
     def initialize(mode, app=nil)
+      raise TypeError, "The second parameter to Session::new should be a rack app if passed." if app && !app.respond_to?(:call)
       @mode = mode
       @app = app
       if Capybara.run_server and @app and driver.needs_server?
@@ -120,9 +123,19 @@ module Capybara
     # Raise errors encountered in the server
     #
     def raise_server_error!
-      raise @server.error if Capybara.raise_server_errors and @server and @server.error
-    ensure
-      @server.reset_error! if @server
+      if @server and @server.error
+        # Force an explanation for the error being raised as the exception cause
+        begin
+          if Capybara.raise_server_errors
+            raise CapybaraError, "Your application server raised an error - It has been raised in your test code because Capybara.raise_server_errors == true"
+          end
+        rescue CapybaraError
+          #needed to get the cause set correctly in JRuby -- otherwise we could just do raise @server.error
+          raise @server.error, @server.error.message, @server.error.backtrace
+        ensure
+          @server.reset_error!
+        end
+      end
     end
 
     ##
@@ -160,7 +173,15 @@ module Capybara
     # @return [String] Path of the current page, without any domain information
     #
     def current_path
-      path = URI.parse(current_url).path
+      # Addressable parsing is more lenient than URI
+      uri = ::Addressable::URI.parse(current_url)
+
+      # If current_url ends up being nil, won't be able to call .path on a NilClass.
+      return nil if uri.nil?
+
+      # Addressable doesn't support opaque URIs - we want nil here
+      return nil if uri.scheme == "about"
+      path = uri.path
       path if path and not path.empty?
     end
 
@@ -256,8 +277,8 @@ module Capybara
     # block, any command to Capybara will be handled as though it were scoped
     # to the given element.
     #
-    #     within(:xpath, '//div[@id="delivery-address"]') do
-    #       fill_in('Street', :with => '12 Main Street')
+    #     within(:xpath, './/div[@id="delivery-address"]') do
+    #       fill_in('Street', with: '12 Main Street')
     #     end
     #
     # Just as with `find`, if multiple elements match the selector given to
@@ -268,13 +289,13 @@ module Capybara
     # assumed to be of the type set in Capybara.default_selector.
     #
     #     within('div#delivery-address') do
-    #       fill_in('Street', :with => '12 Main Street')
+    #       fill_in('Street', with: '12 Main Street')
     #     end
     #
     # Note that a lot of uses of `within` can be replaced more succinctly with
     # chaining:
     #
-    #     find('div#delivery-address').fill_in('Street', :with => '12 Main Street')
+    #     find('div#delivery-address').fill_in('Street', with: '12 Main Street')
     #
     # @overload within(*find_args)
     #   @param (see Capybara::Node::Finders#all)
@@ -293,6 +314,7 @@ module Capybara
         scopes.pop
       end
     end
+    alias_method :within_element, :within
 
     ##
     #
@@ -320,21 +342,77 @@ module Capybara
 
     ##
     #
-    # Execute the given block within the given iframe using given frame name or index.
-    # May be supported by not all drivers. Drivers that support it, may provide additional options.
+    # Switch to the given frame
     #
-    # @overload within_frame(index)
-    #   @param [Integer] index         index of a frame
-    # @overload within_frame(name)
-    #   @param [String] name           name of a frame
+    # If you use this method you are responsible for making sure you switch back to the parent frame when done in the frame changed to.
+    # Capybara::Session#within_frame is preferred over this method and should be used when possible.
+    # May not be supported by all drivers.
     #
-    def within_frame(frame_handle)
-      scopes.push(nil)
-      driver.within_frame(frame_handle) do
-        yield
+    # @overload switch_to_frame(element)
+    #   @param [Capybara::Node::Element]  iframe/frame element to switch to
+    # @overload switch_to_frame(:parent)
+    #   Switch to the parent element
+    # @overload switch_to_frame(:top)
+    #   Switch to the top level document
+    #
+    def switch_to_frame(frame)
+      case frame
+      when Capybara::Node::Element
+        driver.switch_to_frame(frame)
+        scopes.push(:frame)
+      when :parent
+        raise Capybara::ScopeError, "`switch_to_frame(:parent)` cannot be called from inside a descendant frame's "\
+                                    "`within` block." if scopes.last() != :frame
+        scopes.pop
+        driver.switch_to_frame(:parent)
+      when :top
+        idx = scopes.index(:frame)
+        if idx
+          raise Capybara::ScopeError, "`switch_to_frame(:top)` cannot be called from inside a descendant frame's "\
+                                      "`within` block." if scopes.slice(idx..-1).any? {|scope| ![:frame, nil].include?(scope)}
+          scopes.slice!(idx..-1)
+          driver.switch_to_frame(:top)
+        end
+      end
+    end
+
+    ##
+    #
+    # Execute the given block within the given iframe using given frame, frame name/id or index.
+    # May not be supported by all drivers.
+    #
+    # @overload within_frame(element)
+    #   @param [Capybara::Node::Element]  frame element
+    # @overload within_frame([kind = :frame], locator, options = {})
+    #   @param [Symobl] kind      Optional selector type (:css, :xpath, :field, etc.) - Defaults to :frame
+    #   @param [String] locator   The locator for the given selector kind.  For :frame this is the name/id of a frame/iframe element
+    # @overload within_frame(index)
+    #   @param [Integer] index         index of a frame (0 based)
+    def within_frame(*args)
+      frame = _find_frame(*args)
+
+      begin
+        switch_to_frame(frame)
+        begin
+          yield
+        ensure
+          switch_to_frame(:parent)
+        end
+      rescue Capybara::NotSupportedByDriverError
+        # Support older driver frame API for now
+        if driver.respond_to?(:within_frame)
+          begin
+            scopes.push(:frame)
+            driver.within_frame(frame) do
+              yield
+            end
+          ensure
+            scopes.pop
+          end
+        else
+          raise
+        end
       end
-    ensure
-      scopes.pop
     end
 
     ##
@@ -403,7 +481,7 @@ module Capybara
         driver.switch_to_window(window.handle)
         window
       else
-        wait_time = Capybara::Queries::SelectorQuery.new(options).wait
+        wait_time = Capybara::Queries::BaseQuery.wait(options)
         document.synchronize(wait_time, errors: [Capybara::WindowError]) do
           original_window_handle = driver.current_window_handle
           begin
@@ -500,7 +578,7 @@ module Capybara
       old_handles = driver.window_handles
       block.call
 
-      wait_time = Capybara::Queries::SelectorQuery.new(options).wait
+      wait_time = Capybara::Queries::BaseQuery.wait(options)
       document.synchronize(wait_time, errors: [Capybara::WindowError]) do
         opened_handles = (driver.window_handles - old_handles)
         if opened_handles.size != 1
@@ -518,10 +596,16 @@ module Capybara
     # +evaluate_script+ whenever possible.
     #
     # @param [String] script   A string of JavaScript to execute
+    # @param args  Optional arguments that will be passed to the script.  Driver support for this is optional and types of objects supported may differ between drivers
     #
-    def execute_script(script)
+    def execute_script(script, *args)
       @touched = true
-      driver.execute_script(script)
+      if args.empty?
+        driver.execute_script(script)
+      else
+        raise Capybara::NotSupportedByDriverError, "The current driver does not support execute_script arguments" if driver.method(:execute_script).arity == 1
+        driver.execute_script(script, *args.map { |arg| arg.is_a?(Capybara::Node::Element) ?  arg.base : arg} )
+      end
     end
 
     ##
@@ -533,9 +617,15 @@ module Capybara
     # @param  [String] script   A string of JavaScript to evaluate
     # @return [Object]          The result of the evaluated JavaScript (may be driver specific)
     #
-    def evaluate_script(script)
+    def evaluate_script(script, *args)
       @touched = true
-      driver.evaluate_script(script)
+      result = if args.empty?
+        driver.evaluate_script(script)
+      else
+        raise Capybara::NotSupportedByDriverError, "The current driver does not support evaluate_script arguments" if driver.method(:evaluate_script).arity == 1
+        driver.evaluate_script(script, *args.map { |arg| arg.is_a?(Capybara::Node::Element) ?  arg.base : arg} )
+      end
+      element_script_result(result)
     end
 
     ##
@@ -551,11 +641,7 @@ module Capybara
     #   @raise [Capybara::ModalNotFound]  if modal dialog hasn't been found
     #
     def accept_alert(text_or_options=nil, options={}, &blk)
-      text_or_options, options = nil, text_or_options if text_or_options.is_a?(Hash)
-      options[:text] ||= text_or_options unless text_or_options.nil?
-      options[:wait] ||= Capybara.default_max_wait_time
-
-      driver.accept_modal(:alert, options, &blk)
+      accept_modal(:alert, text_or_options, options, &blk)
     end
 
     ##
@@ -565,11 +651,7 @@ module Capybara
     # @macro modal_params
     #
     def accept_confirm(text_or_options=nil, options={}, &blk)
-      text_or_options, options = nil, text_or_options if text_or_options.is_a?(Hash)
-      options[:text] ||= text_or_options unless text_or_options.nil?
-      options[:wait] ||= Capybara.default_max_wait_time
-
-      driver.accept_modal(:confirm, options, &blk)
+      accept_modal(:confirm, text_or_options, options, &blk)
     end
 
     ##
@@ -579,11 +661,7 @@ module Capybara
     # @macro modal_params
     #
     def dismiss_confirm(text_or_options=nil, options={}, &blk)
-      text_or_options, options = nil, text_or_options if text_or_options.is_a?(Hash)
-      options[:text] ||= text_or_options unless text_or_options.nil?
-      options[:wait] ||= Capybara.default_max_wait_time
-
-      driver.dismiss_modal(:confirm, options, &blk)
+      dismiss_modal(:confirm, text_or_options, options, &blk)
     end
 
     ##
@@ -594,11 +672,7 @@ module Capybara
     # @option options [String] :with   Response to provide to the prompt
     #
     def accept_prompt(text_or_options=nil, options={}, &blk)
-      text_or_options, options = nil, text_or_options if text_or_options.is_a?(Hash)
-      options[:text] ||= text_or_options unless text_or_options.nil?
-      options[:wait] ||= Capybara.default_max_wait_time
-
-      driver.accept_modal(:prompt, options, &blk)
+      accept_modal(:prompt, text_or_options, options, &blk)
     end
 
     ##
@@ -608,11 +682,7 @@ module Capybara
     # @macro modal_params
     #
     def dismiss_prompt(text_or_options=nil, options={}, &blk)
-      text_or_options, options = nil, text_or_options if text_or_options.is_a?(Hash)
-      options[:text] ||= text_or_options unless text_or_options.nil?
-      options[:wait] ||= Capybara.default_max_wait_time
-
-      driver.dismiss_modal(:prompt, options, &blk)
+      dismiss_modal(:prompt, text_or_options, options, &blk)
     end
 
     ##
@@ -711,10 +781,27 @@ module Capybara
     end
 
     def current_scope
-      scopes.last || document
+      scope = scopes.last
+      scope = document if [nil, :frame].include? scope
+      scope
     end
 
   private
+    def accept_modal(type, text_or_options, options, &blk)
+      driver.accept_modal(type, modal_options(text_or_options, options), &blk)
+    end
+
+    def dismiss_modal(type, text_or_options, options, &blk)
+      driver.dismiss_modal(type, modal_options(text_or_options, options), &blk)
+    end
+
+    def modal_options(text_or_options, options)
+      text_or_options, options = nil, text_or_options if text_or_options.is_a?(Hash)
+      options[:text] ||= text_or_options unless text_or_options.nil?
+      options[:wait] ||= Capybara.default_max_wait_time
+      options
+    end
+
 
     def open_file(path)
       begin
@@ -738,11 +825,42 @@ module Capybara
 
     def default_fn(extension)
       timestamp = Time.new.strftime("%Y%m%d%H%M%S")
-      path = "capybara-#{timestamp}#{rand(10**10)}.#{extension}"
+      "capybara-#{timestamp}#{rand(10**10)}.#{extension}"
     end
 
     def scopes
       @scopes ||= [nil]
     end
+
+    def element_script_result(arg)
+      case arg
+      when Array
+        arg.map { |e| element_script_result(e) }
+      when Hash
+        arg.each { |k, v| arg[k] = element_script_result(v) }
+      when Capybara::Driver::Node
+        Capybara::Node::Element.new(self, arg, nil, nil)
+      else
+        arg
+      end
+    end
+
+    def _find_frame(*args)
+      within(document) do  # Previous 2.x versions ignored current scope when finding frames - consider changing in 3.0
+        case args[0]
+        when Capybara::Node::Element
+          args[0]
+        when String, Hash
+          find(:frame, *args)
+        when Symbol
+          find(*args)
+        when Integer
+          idx = args[0]
+          all(:frame, minimum: idx+1)[idx]
+        else
+          raise TypeError
+        end
+      end
+    end
   end
 end
